{% if persona_definition %}
- name: "persona"
  raw_string: |
    {{ settings.start_token }}{{ canonicalize_user_name(username) }}: {{ escape_sequences(persona_definition) }}{{ settings.eom }}
  {% set settings.start_token = settings.bom %}
{% endif %}
